(* Time-stamp: <modified the 17/10/2019 (at 21:09) by Erwan Jahier> *)

open Sasacore
(* Process programmer API *)
type action = string (* just a label *)

type 's neighbor = {
  state:  's ;
  pid: unit -> string; 
  spid:  unit -> string;
  reply: unit -> int; 
  weight: unit -> int; 
}

exception Not_available

(** processes local state (user defined) *) 
let (state : 's neighbor -> 's) = fun s -> s.state

(** This pid is not available in all simulation modes (e.g.,
   anonymous) *)
let (pid : 's neighbor -> string) = fun s -> s.pid ()
        

(* spid (self-pid) of the process this neighbor is the neighbor of *)
let (spid : 's neighbor -> string) = fun s -> s.spid ()
            
(** Returns the channel number that let this neighbor access to the
   content of the process, if it neighbor can access it.  Returns -1
   if the neigbor can not access to the process, which may happen in
   directed graphs only.  This info is not available in all
   simulation modes. *)
let (reply : 's neighbor -> int) = fun s -> s.reply ()
                
(** Makes sense in directed graphs only *)
let (weight : 's neighbor -> int) = fun s -> s.weight ()
                    

type 's enable_fun = 's -> 's neighbor list -> action list
type 's step_fun   = 's -> 's neighbor list -> action -> 's
type 's state_init_fun = int -> 's

type 's algo_to_register = {
  algo_id: string;
  init_state: int -> 's;
  enab: 's enable_fun;
  step: 's step_fun;
  actions: action list option
}
type 's to_register = {
  algo : 's algo_to_register list;
  state_to_string: 's -> string;
  state_of_string: (string -> 's) option;
  copy_state: 's -> 's;
}

let (to_reg_neigbor : 's Register.neighbor -> 's neighbor) =
  fun n ->
    {
      state = n.Register.state ;
      pid = n.Register.pid;
      spid = n.Register.pid;
      reply = n.Register.reply; 
      weight = fun () -> match n.Register.weight () with
          None -> raise Not_available
        | Some w -> w; 
    } 

let (to_reg_enable_fun : 's enable_fun ->
     's Register.neighbor list -> 's -> action list) =
  fun f nl s ->
    f s (List.map to_reg_neigbor nl)
  
let (to_reg_step_fun : 's step_fun ->
     's Register.neighbor list -> 's -> action -> 's) =
  fun f nl s a ->
    f s (List.map to_reg_neigbor nl) a
  

let (register1 : 's algo_to_register -> unit) =
  fun s ->
    Register.reg_enable     s.algo_id (to_reg_enable_fun s.enab);
    Register.reg_step       s.algo_id (to_reg_step_fun s.step);
    Register.reg_init_state s.algo_id s.init_state;
    (match s.actions with None -> () | Some al -> Register.reg_actions s.algo_id al);
    ()
      
let registered = ref false
(* exported *)
let (register : 's to_register -> unit) =
  fun s ->
    if !registered then failwith "Register can only be called once!";
    registered := true;
    Register.reg_value_to_string s.state_to_string;
    Register.reg_copy_value s.copy_state;
    List.iter register1 s.algo;
    (match s.state_of_string with None -> () | Some f -> Register.reg_value_of_string f);
    ()


let card = Register.card

let get_graph_attribute = Register.get_graph_attribute                           
let min_degree = Register.min_degree
let mean_degree = Register.mean_degree
let max_degree = Register.max_degree
let is_cyclic = Register.is_cyclic
let is_connected = Register.is_connected
let is_directed = Register.is_directed
let is_tree = Register.is_tree
let height = Register.height
let links_number = Register.links_number
let diameter = Register.diameter
