(* Time-stamp: <modified the 23/10/2019 (at 13:49) by Erwan Jahier> *)
(** {1 The Algorithm programming Interface.} *)
(** 
    {1 What's need to be provided by users.}
    A SASA process is an instance of an algorithm defined via this
    module.   

    In order to define a SASA algorithm, one need to provide an enable
    and a step function:
   - the enable function states which actions can be triggered;
   - the step function triggers an enable action.

 The enable and the step functions should be of the  type
 [ 's enable_fun ] and [ 's step_fun ] respectively.
 *)

type 's neighbor
type action  = string 
type 's enable_fun = 's -> 's neighbor list -> action list
type 's step_fun   = 's -> 's neighbor list -> action -> 's
(** The first argument  holds the current state of the  process.  As it is
    polymorphic (['s]),  algorithms designers  can put anything  they need
    into this state (an integer,  a structure, etc.).  The only constraint
    is that all algorithms should use the same type.
    
    The  second  argument  holds  the  process  neighbors. Note that  SASA
    processes,  that  live  in  graph  nodes, can  only  access  to  their
    immediate  neighbors.  From  each neighbor,  a process  can access  to
    various information thanks to the functions below.
*)

type 's state_init_fun = int -> 's
(**  The  initial  value of  the  local  state  can  be set  using  an
   initialization function that takes as unique argument the number of
   neighbors of the node.  *)

(**  When a neighbor information can not be accessed (e.g., the pid
     anonymous networks) the Not_available exception is raised.  *)
exception Not_available
  
(** Returns the processes local state *) 
val state : 's neighbor -> 's

(** Returns the process id of the current process neighbor. This info
   is not available in all simulation modes. *)
val pid : 's neighbor -> string

(** Returns the process id of the current process

   “To see ourselves as others see us is a most salutary gift. Hardly
   less important is the capacity to see others as they see
   themselves.”  ― Aldous Huxley, The Doors of Perception.
  *)
val spid : 's neighbor -> string

(** Returns  the channel  number  that  let this  neighbor  access to  the
    content of the  process, if it neighbor can access  it.  Returns -1 if
    the  neigbor can  not  access  to the  process,  which  may happen  in
    directed graphs  only.  This info  is not available in  all simulation
    modes.
 *)
val reply : 's neighbor -> int

(** Returns the weight decoration the current node and it neighbor
   (which makes sense in directed graphs only) *)
val weight : 's neighbor -> int 


(** {1 Access to Topological Information}

   When  writing  algorithms, one  can  also  have acces  to  topological
   information (i.e., information that are relative to the graph) *)
  
val card : unit -> int
val min_degree : unit -> int
val mean_degree : unit -> float
val max_degree : unit -> int
val is_directed : unit -> bool
val is_cyclic : unit -> bool
val is_connected : unit -> bool
val is_tree : unit -> bool
val height : unit -> (string (* the node id *) -> int) option
val links_number : unit -> int
val diameter : unit -> int

(** It is possible to set some global parameters in the dot file
   using graph attributes. This function allows one the get their
   values. *)
val get_graph_attribute : string -> string


(** {1 Code Registration}

The [register: 's to_register -> unit] function must be called once in
   the user code.  *)

type 's algo_to_register = {
  algo_id   : string;
  init_state: 's state_init_fun;
  enab      : 's enable_fun;
  step      : 's step_fun;
  actions   : action list option (** Mandatory in custom demon mode *)
}
type 's to_register = {
    algo : 's algo_to_register list;
    state_to_string:  's  -> string;
    state_of_string: (string -> 's) option;
    copy_state: 's -> 's;
  }
(** -  For the [state_to_string] field,  the idea is to  print the raw
   values contained in ['s].  If a  value is omitted, one won't see it
   in the simulation outputs.  If  one prepend a value with "some_id",
   some_id will we  used in the simulation outputs.   Otherwise, an id
   will be invented by sasa.

  - For  the [state_of_string]  field, if  some function  is provided,
   sasa should be able to parse state init values in the dot.

*)


(** To be called once *)
val register : 's to_register -> unit
  
(** {1 Automatic Generation of Registration Code}

   Once enable and step functions are defined, they need to be
   registred by calling the register function.

    An alternative to writing this registration code is to let sasa 
    generate it with the "--gen-register" (or "-reg") option.
    In this case, one needs to follow some naming conventions (w.r.t file 
    and function names):

    (1) The internal state (i.e. the 's shared by all algos) should be 
    defined in a file named "state.ml" that defines the following 4 entities:
     {[
       type t = define_me
       let (to_string: t -> string) = fun _ -> "define_me"
       let (of_string: string -> t option) = fun _ -> None
       let (copy : t -> t) = fun x -> x
     ]}

    If one does not want/need to provide the of_string state parser, returning
    None is fine. This is mandatory only if one wants to define initial values
    in the dot file. 

    Defining a copy that is not the identity is necessary if the state is not 
    functionnal (e.g., if it contains an array or an Hashtbl).

    In the file "state.ml" does not exist in the current directory, a skeleton 
    is generated.

    (2) All the algos mentionned in the dot file should define the 
    following functions:
     {[
       let (init_state: int -> State.t) = xxx
       let (enable_f: State.t neighbor list -> State.t -> action list) = xxx
       let (step_f : State.t neighbor list -> State.t -> action -> State.t ) = xxx
       let actions = Some ["action1";"action2"];
     ]}
*)
