(* Time-stamp: <modified the 04/11/2019 (at 10:47) by Erwan Jahier> *)


type t = {
  mutable topo: string;
  mutable length: int;
  mutable verbose: int;
  mutable demon: Demon.t;
  mutable rif: bool;
  mutable no_data_file: bool;
  mutable quiet: bool;
  mutable seed: int option;
  mutable replay_seed: bool;
  mutable ifi: bool;
  mutable gen_lutin: bool;
  mutable gen_oracle: bool;
  mutable dummy_input: bool;
  mutable output_algos: bool;
  mutable gen_register: bool;
  
  mutable _args : (string * Arg.spec * string) list;
  mutable _user_man  : (string * string list) list; 
  mutable _hidden_man: (string * string list) list; 

  mutable _others : string list;
  mutable _margin : int;
}

let usage_msg tool = ("usage: " ^ tool ^ " [<option>]* <topology>.dot 
use -h to see the available options.
" )
let print_usage tool = Printf.printf "%s\n" (usage_msg tool); flush stdout


let (make_args : unit -> t) = 
  fun () -> 
    {
      topo = "";
      length = 10000;
      verbose = 0;
      demon = Demon.Distributed;
      rif = false;
      no_data_file = false;
      quiet = false;
      seed = None;
      replay_seed = false;
      ifi = false;
      gen_lutin = false;
      gen_oracle = false;
      dummy_input = false;
      output_algos = false;
      gen_register = false;
      _args = [];        
      _user_man  = [];   
      _hidden_man  = []; 
      _others = [];
      _margin =12;
    }


let pspec args os  (c, ml) = (
  let (m1, oth) = match ml with
	 |	h::t -> (h,t)
	 |	_ -> ("",[])
  in
  let t2 = String.make args._margin ' ' in
  let cl = String.length c in
  let t1 = if (cl < args._margin ) then
	 String.make (args._margin - cl) ' '
  else
	 "\n"^t2
  in
	 Printf.fprintf os "%s%s%s" c t1 m1;
	 List.iter (function x -> Printf.fprintf os "\n%s%s" t2 x) oth ;
	 Printf.fprintf os "\n" ;
)

let options args oc = (
	let l = List.rev args._user_man in
	List.iter (pspec args oc) l
)
let more_options args oc = (
	let l = List.rev (args._hidden_man) in
	List.iter (pspec args oc) l
)
let (mkopt : t -> string list -> ?hide:bool -> ?arg:string -> Arg.spec ->
     string list -> unit) =
  fun opt ol ?(hide=false) ?(arg="") se ml ->
    let treto o = opt._args <- (o, se, "")::opt._args in
	   List.iter treto ol ;
	   let col1 = (String.concat ", " ol)^arg in
	     if hide 
        then opt._hidden_man <- (col1, ml)::opt._hidden_man
	     else opt._user_man   <- (col1, ml)::opt._user_man

let myexit i = exit i
(*******************************************************************************)
(* seeds stuff *)

let seed_set args s = 
  (match s with
   | Some i ->
     Printf.fprintf stderr "The sasa random engine seed is set to %i\n" i;
     Random.init i;
     flush stderr;
   | None -> ()
  );
  args.seed <- s

let seed_file_name args =
  Printf.sprintf "sasa-%s.seed"  args.topo 

(* for --replay *)
let reset_the_seed_to_last args =
  let f = seed_file_name args in
  try
    let ic = open_in f in
    let seed = int_of_string (input_line ic) in
    args.seed <- Some seed;
    Printf.eprintf "Replay the sasa run using the seed in %s\n" f;
    flush stderr;
    true
  with _ ->
    Printf.eprintf "W: cannot recover the seed in %s\n" f;
    flush stderr;
    false
              
let rec seed_get args = match args.seed with
  | Some i -> i
  | None ->
     (* No seed is set: 
         - in -replay mode, we first try to read the seed in the seed file
         - otherwise, we create a random seed and save if into args, and 
           into a file for -replay *)
     if args.replay_seed && reset_the_seed_to_last args then (seed_get args) else
       let seed = Random.self_init (); Random.int 1073741823 in
       let seed_file = seed_file_name args in
       let oc = open_out seed_file in
       Printf.fprintf oc "%d\n%s\n" seed (Mypervasives.entete "#" SasaVersion.str SasaVersion.sha);
       flush oc;
       close_out oc;
       seed_set args (Some seed);
       seed
       
(*******************************************************************************)

(*** User Options Tab **)
let (mkoptab : string array -> t -> unit) = 
  fun argv args ->  
    (
    mkopt args  ["--synchronous-demon";"-sd"]
      (Arg.Unit(fun () -> args.demon <- Demon.Synchronous))
      ["Use a Synchronous demon"];

    mkopt args  ["--central-demon";"-cd"]
      (Arg.Unit(fun () -> args.demon <- Demon.Central))
      ["Use a Central demon (selects exactly one action)"];
    
    mkopt args  ["--locally-central-demon";"-lcd"]
      (Arg.Unit(fun () -> args.demon <- Demon.LocallyCentral))
      ["Use a Locally Central demon";
       "(i.e., never activates two neighbors actions in the same step)"];
    mkopt args  ["--distributed-demon";"-dd"]
      (Arg.Unit(fun () -> args.demon <- Demon.Distributed))
      ["Use a Distributed demon (which select at least one action). This is the default demon."];
    
    mkopt args  ["--custom-demon";"-custd"]
      (Arg.Unit(fun () -> args.demon <- Demon.Custom;args.rif <- true))
      ["Use a Custom demon (forces --rif)"];

   mkopt args ~hide:true ["--rif";"-rif"]
      (Arg.Unit(fun () -> args.rif <- true))
      ["Display only outputs on stdout (i.e., behave as a rif input file)"];

   mkopt args  ~hide:true  ["--no-data-file";"-nd"]
      (Arg.Unit(fun () -> args.no_data_file <- true))
      ["Do not generate any data file"];

    mkopt args  ["--seed";"-seed"]
      (Arg.Int(fun i -> seed_set args (Some i)))
      ["Set the pseudo-random generator seed of build-in demons (wins over --replay)"];
    
    mkopt args  ["--replay";"-replay"]
      (Arg.Unit(fun () -> args.replay_seed <- true))
      ["Use the last generated seed to replay the last run"];

   mkopt args ~hide:true ["--gen-lutin-demon";"-gld"]
      (Arg.Unit(fun () -> args.gen_lutin <- true))
      ["Generate Lutin demons and exit"];

   mkopt args ~hide:true ["--gen-lustre-oracle-skeleton";"-glos"]
      (Arg.Unit(fun () -> args.gen_oracle <- true))
      ["Generate a Lustre oracle skeleton"];

   mkopt args ~hide:true ["--list-algos";"-algo"]
     (Arg.Unit(fun () -> args.output_algos <- true))
      ["Output the algo files used in the dot file and exit. "];
    
   mkopt args ~hide:false ["--gen-register";"-reg"]
     (Arg.Unit(fun () -> args.gen_register <- true))
      ["Generates the registering file and exit. "];
    
   mkopt args ~hide:true ["--dummy-input"]
      (Arg.Unit(fun () -> args.dummy_input <- true))
      ["Add a dummy input to sasa so that built-in demon can be used from rdbg"];

    mkopt args ~hide:true ["--ignore-first-inputs"; "-ifi"]
      (Arg.Unit(fun () -> args.ifi <- true))
      ["Ignore first inputs (necessary to use luciole via ";
       "lurette/rdbg/luciole-rif)"];

    mkopt args ["--length";"-l"] ~arg:" <int>"
      (Arg.Int (fun i -> args.length <- i))
      ["Maximum number of steps to be done (" ^
       (string_of_int args.length) ^ " by default).\n"];

    mkopt args ["--version";"-version";"-v"]
      (Arg.Unit (fun _ ->
           (print_string (SasaVersion.str^"-"^SasaVersion.sha^"\n");
            flush stdout;exit 0)))
      ["Display the sasa version and exit."];

    mkopt args ~hide:true ["--ocaml-version"]
      (Arg.Unit (fun _ -> (print_string (Sys.ocaml_version^"\n"); flush stdout; exit 0)))
      ["Display the version ocaml version sasa was compiled with and exit."];

    mkopt args  ["--verbose";"-vl"] ~arg:" <int>"
      (Arg.Int (fun i -> args.verbose <- i))   ["Set the verbose level"];

    mkopt args ["--help";"-help"; "-h"]
      (Arg.Unit (fun _ -> print_usage (argv.(0)); options args stdout; exit 0))
      ["Display main options"];

    mkopt args ["--more";"-m"] (Arg.Unit (fun () -> more_options args stdout; exit 0)) 
      ["Display more options"]

  )

(* all unrecognized options are accumulated *)
let (add_other : t -> string -> unit) =
  fun opt s -> 
    opt._others <- s::opt._others

let current = ref 0;;
let first_line b = (
	try (
		let f = String.index b '\n' in
		String.sub b 0 f
	) with Not_found -> b
)
let file_notfound f = (
	prerr_string ("File not found: \""^f^"\"");
	prerr_newline ();
	myexit 1
)
let unexpected s = (
	prerr_string ("unexpected argument \""^s^"\"");
	prerr_newline ();
	myexit 1
)

let parse argv = (
  let save_current = !current in
  try (
    let args = make_args () in 
    mkoptab argv args;
	 Arg.parse_argv ~current:current argv args._args (add_other args) (usage_msg argv.(0));
    (List.iter 
       (fun f -> 
         if (String.sub f 0 1 = "-") then
           unexpected f 
         else if not (Sys.file_exists f) then
           file_notfound f
         else ()
       ) 
       args._others);
    current := save_current;
    args.topo <- (match args._others with
          [] -> 
          Printf.fprintf stderr "*** The topology file is missing in '%s'\n%s\n"
            (argv.(0)) (usage_msg argv.(0));  
          exit 2; 
        | x::_ -> x
      );
    args
  )
  with
  | Arg.Bad msg ->
    Printf.fprintf stderr "*** Error when calling '%s': %s\n%s\n" (argv.(0))
      (first_line msg) (usage_msg argv.(0)); exit 2; 
  | Arg.Help msg -> 
    Printf.fprintf stdout "%s\n%s\n" msg (usage_msg argv.(0)); 
    exit 0
)
