(* Time-stamp: <modified the 15/11/2019 (at 11:10) by Erwan Jahier> *)

open Graph
open Graph.Dot_ast

type node_id = string
type node = {
  id: node_id;
  file: string;
  init: string; 
}


type t = {
  nodes: node list;
  succ: node_id -> (int option * node_id) list;
  of_id: node_id -> node;
  directed:bool
}

type node_info_t = (string, node) Hashtbl.t
let node_info:node_info_t = Hashtbl.create 100

type node_succ_t = (string, (int option * node_id) list) Hashtbl.t
let node_succ:node_succ_t = Hashtbl.create 100
        
let (of_id:Dot_ast.id -> string) =
  function Ident str | Html str | Number str | String str -> str

let (of_node_id:Dot_ast.node_id -> string) = fun id -> of_id (fst id)
let (of_node:Dot_ast.node -> string) = function
  | NodeId node_id -> of_node_id node_id
  | NodeSub _subgraph -> failwith "subgraphs are not supported in sasa "


let (get_file: Dot_ast.node_id  -> Dot_ast.attr list -> string) =
  fun node_id attrs ->
    let attrs = List.flatten attrs in (* XXX why a list of list ? *)
    try  (
        match List.assoc (Ident "algo") attrs with
          None -> assert false
        | Some id -> of_id id
      ) 
    with Not_found -> failwith ((of_node_id node_id)^" should have an algo attribute")

let (get_init: Dot_ast.attr list -> string) =
  fun attrs ->
    let attrs = List.flatten attrs in (* XXX why a list of list ? *)
    let init_list =
      List.fold_left
        (fun acc (id,idopt) ->
           if id <> Ident "init" then acc else
             match idopt with
             | Some (String id) -> id 
             | _ -> acc
        )
        ""
        attrs
    in
    init_list

let rec (get_weight: Dot_ast.attr -> int option ) =
  function
  | (Ident "weight", Some Number n)::_ -> (try Some (int_of_string n) with _  -> None)
  | _::tail -> get_weight tail
  | [] -> None

let (do_graph_attr : (Dot_ast.id * Dot_ast.id option) list -> unit) =
  fun l -> 
    let f  = function
      | id, Some v -> Register.set_graph_attribute (of_id id) (of_id v)
      | _, None -> ()
    in
    List.iter f l
      
let (do_stmt: bool -> node list -> Dot_ast.stmt -> node list) =
  fun directed n stmt ->
    match stmt with
    | Node_stmt (node_id, attrs) ->
      let id = of_node_id node_id in
      let inits = get_init attrs  in
      let node = { id=id ; file = get_file node_id attrs ; init = inits } in
      Hashtbl.replace node_info id node;
      node::n
    | Edge_stmt (dot_node, nodes, attrs) ->
      let node = of_node dot_node in
      let nodes = List.map of_node nodes in
        (* for egdes written "a -- b -- c -- d", which
           is a shortcut for "
           a -- b
           b -- c
           c -- d
           "
           Graph.Dot.parse_dot_ast returns the pair a,[b,c,d]
           which is weird IMHO.   
           The code below add the missing edges:
        *)
      let attrs = List.flatten attrs in (* XXX why a list of list ? *)
      let weight = if directed then get_weight attrs else None in
      let add_edge n1 n2 =
        if n1 = n2 then failwith
            (Printf.sprintf
               "Bad topology: %s can not ne a neighbor of itself!" n1);
        let pn1 = try Hashtbl.find node_succ n1 with Not_found -> [] in
        let pn2 = try Hashtbl.find node_succ n2 with Not_found -> [] in
        if not (List.mem (weight,n2) pn1) then
          Hashtbl.replace node_succ n1 ((weight,n2)::pn1);
        if not directed && not (List.mem (None,n1) pn2) then
          Hashtbl.replace node_succ n2  ((None,n1)::pn2);
        n2
      in
      ignore (List.fold_left add_edge node nodes);
      n
    | Attr_graph attrs -> do_graph_attr (List.flatten attrs); n
    | Attr_node  _attrs -> n 
    | Attr_edge  _attrs ->  n
    | Equal (_id1, _id2) -> assert false
    | Subgraph _subgraph -> assert false


let (read: string -> t) = fun f ->
  let dot_file = Graph.Dot.parse_dot_ast f in
  assert (not dot_file.strict);
  let res = List.fold_left (do_stmt dot_file.digraph) [] dot_file.stmts in
  {
    nodes = List.rev res;
    succ = (fun str -> try Hashtbl.find node_succ str with Not_found -> []);
    of_id = (fun str -> try Hashtbl.find node_info str with Not_found ->
                          failwith (str^ " unknown node id")
            );
    directed = dot_file.digraph
  }

let (to_adjency: t -> bool array array) =
  fun t ->
    let n = List.length t.nodes in
    let rank_node_tbl = Hashtbl.create n in
    let m = Array.make_matrix n n false in
    let rank_node = Hashtbl.find rank_node_tbl in
    List.iteri (fun i n -> Hashtbl.add rank_node_tbl n.id i) t.nodes;
    List.iteri
      (fun i n  ->
         List.iter (fun (_,target) -> m.(i).(rank_node target) <- true) (t.succ n.id)
      )
      t.nodes;
    m
let (get_degree:t -> int*int) =
  fun t -> if t.nodes = [] then 0,0
    else
      let node_deg n = List.length (t.succ (n.id)) in
      let d_start = node_deg ((List.hd t.nodes)) in
      List.fold_left (fun (d_min,d_max) n ->
          (min (node_deg n) d_min, max (node_deg n) d_max)
      ) (d_start,d_start) (List.tl t.nodes)

let (get_nb_link: t -> int) =
  fun t ->
  let res = List.fold_left
      (fun acc n -> ((List.length (t.succ n.id))) + acc) 0 t.nodes
  in 
  if t.directed then res else res/2
    
let (get_mean_degree : t -> float) = 
  fun t -> 
    (float_of_int (get_nb_link t)) /. (float_of_int (List.length t.nodes))

let bfs : (t -> string -> bool * string list) = 
  fun t n ->
    let q = Queue.create () in
    let discovered = ref [n] and parent = ref (function _ -> "") in
    let cyclic = ref false in
    Queue.add n q;
    while not (Queue.is_empty q) do
      let node = Queue.take q in
      parent := List.fold_left (fun parents (_,suc) -> 
                  if List.for_all (fun disc -> disc <> suc) !discovered 
                  then (
                    Queue.add suc q;
                    discovered := (suc)::!discovered;
                    function a -> if a = suc then node else parents a
                  ) else ((
                    if suc <> (parents node) 
                    then 
                      cyclic := true);
                    parents
                  )
                ) !parent (t.succ node)

    done;
    (!cyclic, !discovered)

let is_connected_and_cyclic : t -> bool*bool = 
  fun t -> match t.nodes with
    | [] -> (false,false)
    | hd::_ ->
      let (cyclic,bfs_nodes) = (bfs t hd.id) in
      ((List.compare_lengths t.nodes bfs_nodes) = 0, cyclic)

let rec height : string list -> t -> string -> int = 
  fun parents t n ->
    (List.fold_left (fun h (_,succ) -> 
         if List.mem succ parents then
           h
         else
           max h (height (n::parents) t succ)) (-1) (t.succ n)) + 1

let get_height : t -> string -> int = 
  fun t -> 
    height ([]) t

